////////////////////////////////////////////////////////////
//	Public functions
////////////////////////////////////////////////////////////

// Global variable
var g_objRegistry = null;
var g_nRegistryType = 0;
var g_strSplitter = "#@$@#";
var g_thunderComponent = null;

// Registry access
// root key values:
// 0: HKEY_CLASSES_ROOT = 0x80000000
// 1: HKEY_CURRENT_CONFIG = 0x80000005
// 2: HKEY_CURRENT_USER = 0x80000001
// 3: HKEY_LOCAL_MACHINE = 0x80000002
// 4: HKEY_USERS = 0x80000003
function createRegistryObject()
{
	if(g_objRegistry)
		return true;
	
	if ("@mozilla.org/windows-registry-key;1" in Components.classes)
	{
		g_objRegistry = Components.classes["@mozilla.org/windows-registry-key;1"].createInstance(Components.interfaces.nsIWindowsRegKey);
		g_nRegistryType = 2;
	}
	else if("@mozilla.org/winhooks;1" in Components.classes)
	{
		g_objRegistry = Components.classes["@mozilla.org/winhooks;1"].getService(Components.interfaces.nsIWindowsRegistry);
		g_nRegistryType = 1;
	}
	else if("@mozilla.org/browser/shell-service;1" in Components.classes)
	{
		g_objRegistry = Components.classes["@mozilla.org/browser/shell-service;1"].getService(Components.interfaces.nsIWindowsShellService);
		g_nRegistryType = 1;
	}
	else
	{
		g_objRegistry = null;
		g_nRegistryType = 0;
	}
	
	if(g_nRegistryType == 0)
		return false;
	return true;
}

function regReadString(strPath, strName)
{
	if(!createRegistryObject())
		return false;
	if(g_nRegistryType == 1)
	{
		return g_objRegistry.getRegistryEntry(3, strPath, strName);
	}
	else if(g_nRegistryType == 2)
	{
		g_objRegistry.open(Components.interfaces.nsIWindowsRegKey.ROOT_KEY_LOCAL_MACHINE,
						   strPath, Components.interfaces.nsIWindowsRegKey.ACCESS_READ);
		return g_objRegistry.readStringValue(strName);	
	}
}

function regWriteString(strPath, strName, strString)
{
	if(!createRegistryObject())
		return false;

	if(g_nRegistryType == 1)
	{
		return g_objRegistry.setRegistryEntry(3, strPath, strName, strString, strString.length);
	}
	else if(g_nRegistryType == 2)
	{
		g_objRegistry.open(ROOT_KEY_LOCAL_MACHINE, strPath, Components.interfaces.nsIWindowsRegKey.ACCESS_WRITE);
		return g_objRegistry.writeStringValue(strName, strString);	
	}
}

// Unicode string convert
function fromUnicodeString(aCharset, strUnicode)
{
	var unicodeConverter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
	unicodeConverter.charset = aCharset;
	strUnicode = unicodeConverter.ConvertFromUnicode(strUnicode);
	return strUnicode;
}

// File access
function isFileExists(strPath)
{
	var objFile = Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);
	objFile.initWithPath(strPath);
	try
	{
		objFile.isFile();
		return true;
	}
	catch(e)
	{
		return false;
	}
}

function readFile(strPath)
{
	var objFile = Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);
	objFile.initWithPath(strPath);
	
	var objInputStream = Components.classes["@mozilla.org/network/file-input-stream;1"].createInstance(Components.interfaces.nsIFileInputStream);
	var objSIStream = Components.classes["@mozilla.org/scriptableinputstream;1"].createInstance(Components.interfaces.nsIScriptableInputStream);
	
	objInputStream.init(objFile, 1, 0, false);
	objSIStream.init(objInputStream);
		
	var data = new String();
	data += objSIStream.read(-1);
	
	objSIStream.close();
	objInputStream.close();
	return data;
}


function writeFile(strPath, strString)
{
	var objFile = Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);
	objFile.initWithPath(strPath);
	
	var objOutStream = Components.classes["@mozilla.org/network/file-output-stream;1"].createInstance(Components.interfaces.nsIFileOutputStream);
	var nFlags = 0x02 | 0x08 | 0x20; // wronly | create | truncate
	objOutStream.init(objFile, nFlags, 0664, 0);
	objOutStream.write(strString, strString.length);
	objOutStream.close();
	return true;
}


// Execute file
function runExe(strFile, strArg)
{
	var objFile = Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);
	objFile.initWithPath(strFile);
	
	if(objFile.isFile() && objFile.isExecutable())
	{
		var objProcess = Components.classes["@mozilla.org/process/util;1"].createInstance(Components.interfaces.nsIProcess);
		objProcess.init(objFile);

		var args = new Array();
		args[args.length] = strArg;
		objProcess.run(false, args, args.length);
	}
}

// call thunder
function callThunder(strUrls, strAsync)
{
	createThunderComponent();
	g_thunderComponent.CallThunder(strUrls, strAsync);
}


